/*
 * Copyright (C) 2014  Filippo Turato
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>
 *
 */

/*
MIDI CONTROLLER FIRMWARE - Versione 0.1

- Invia comandi MIDI program change
- midi channel fisso a 1

TODO:

- migliore implementazione dei pulsanti
- menu` di configurazione che compare se all'accensione e` premuto il tasto 0.
      - imposta midi channel
      - da un nome/descrizione ai preset 

*/


#include <SPI.h>
#include <LiquidCrystal.h>
#include <midi_Namespace.h>
#include <MIDI.h>
#include <midi_Defs.h>
#include <midi_Message.h>
#include <midi_Settings.h>
//#include <avr/sleep.h>      


const byte REGISTER_READ = B01000001; 
const byte REGISTER_WRITE = B01000000; 


/* INDIRIZZI DEI REGISTRI DELL'IO EXPANDER MCP23S18 */

#define		IOCON		0x0A		// REGISTRO DI CONFIGURAZIONE

#define		IODIRA		0x00		// REGISTRO I/O DIRECTION
#define		IODIRB		0x01		// 1=input

#define		IPOLA		0x02		// POLARITA` I/O
#define		IPOLB		0x03		// 1= Inverted

#define		GPIOA		0x12		// INDIRIZZO PORTA GP I/O
#define		GPIOB		0x13		// 

#define		GPINTENA	0x04		// ABILITA IOC (interrupt on change) 
#define		GPINTENB	0x05		// 1 = enabled 

#define		INTCONA		0x08		// INTERRUPT CONTROL REGISTER
#define		INTCONB		0x09		// 1 = INT IF INPUT DIFFER FROM DEFVAL  0=INTERRUPT ON PIN CHANGE

#define		DEFVALA		0x06		// MCP23017 IOC Default value
#define		DEFVALB		0x07		// MCP23017 if INTCONA set then INT. if diff. 

#define		GPPUA		0x0C		// INTERNAL PULLUPS
#define		GPPUB		0x0D		// 1= PULLUP ABILITATO 

#define		OLATA		0x14            // ACCESS OUTPUT LATCHES    
#define		OLATB		0x15            //

#define		INTFA		0x0E            // INTERRUPT FLAG REGISTER
#define		INTFB		0x0F            // 1 = PIN CAUSED INTERRUPT

#define		INTCAPA         0x10            // INTERRUPT CAPTURE VALUE
#define		INTCAPB		0x11            // GPIO PORT STATUS CAPTURED ON INTERRUPT

const byte IOCON_BANK   =  B10000000; // se impostato i regiostri di ogni porta sono separati, altrimenti gli indirizzi sono sequenziali
const byte IOCON_MIRROR =  B01000000; // se impostato i pin INT sono internamente connessi in configurazione OR, altrimenti sono indipendenti
const byte IOCON_SEQOP  =  B00100000; // disabilita l'incremento sequenziale dell'indirizzo dei registri
const byte IOCON_ODR    =  B00000100; // INT diventa open drain se impostato
const byte IOCON_INPOL  =  B00000010; // se impostato INT e` in configurazione ACTIVE HIGH, altrimenti e` il contrario
const byte IOCON_INTCC  =  B00000001; // se impostato per cancellare l'input bisogna leggere il registro INTCAP, altrimenti GPIO

//pin per SPI e I/O expander
const byte pinCS  = 10;
const byte pinCSLED  = A0;
const byte pinSCK = 13;
const byte pinSI  = 11;
const byte pinSO  = 12;
const byte pinRST = 8;
//pin di interrupt
const byte pinINT = 2;

const int wait_time = 500;
unsigned long int time = 0;

//Inizializzazione della libreria LCD: LiquidCrystal(rs, enable, d4, d5, d6, d7) 
//LiquidCrystal lcd(9, 3, 4, 5, 6, 7);
LiquidCrystal lcd(9, 3, 4, 5, 6, 7);

volatile boolean interruptChange = false;

byte ioexp[2]={0,0};
byte program=0;
byte pgm_dec = 0;
byte pgm_uni = 0;
word *data = NULL;
struct gpio {
  byte a;
  byte b;
};
byte midichannel = 1;
MIDI_CREATE_DEFAULT_INSTANCE();




// Routine di interrupt
void hoy() {
   interruptChange = true;
//   sleep_disable ();         // first thing after waking from sleep
   detachInterrupt (0);        // stacca l'interrupt finche` non e` stato elaborato
}

// calcola il logaritmo in base binaria
byte log2(word v) {
 
  byte r = 0;
 
  while (v >>= 1)
  {
    r++;
  }
  
  return r;

}

//conta il numero di bit impostati a 1
byte popCount(word v)
{
  byte c;
  for (c = 0; v; c++) 
  {
    v &= v - 1; // clear the least significant bit set
  }
 
  return c;
}

void setLed(byte n)
{
  word data = 1;
  data = ~(data << n);
  struct gpio *lights;
  lights = (gpio *) &data;
  digitalWrite(pinCSLED, LOW);
  SPI.transfer(REGISTER_WRITE);
  SPI.transfer(GPIOA);
  SPI.transfer(lights->a); 
  SPI.transfer(lights->b);
  digitalWrite(pinCSLED, HIGH);
}


void setup() {

  pinMode(pinCS, OUTPUT);
  pinMode(pinCSLED, OUTPUT);
  pinMode(pinSCK,OUTPUT);
  pinMode(pinSI, OUTPUT);
  pinMode(pinSO,  INPUT);
  pinMode(pinINT, INPUT);
  pinMode(pinRST, OUTPUT);
  
  //16 caratteri 2 colonne
  lcd.begin(16, 2);
  // Print a message to the LCD.
  lcd.setCursor(0,0);
  lcd.print("Inizializzazione");
  lcd.setCursor(0,1);  
  lcd.print("in corso...");
  
  digitalWrite(pinCS, HIGH);
  digitalWrite(pinCSLED, HIGH);
  // Inizializziamo l'IO expander
  digitalWrite(pinRST, LOW);
  delay(1000);
  digitalWrite(pinRST, HIGH);
  delay(100);
  SPI.begin();
  SPI.setBitOrder(MSBFIRST);
  SPI.setClockDivider(SPI_CLOCK_DIV2);
  delay(1000);
  
  // l'io expander per i led
  //output per i led
  digitalWrite(pinCSLED, LOW);
  SPI.transfer(REGISTER_WRITE);
  SPI.transfer(IODIRA);
  SPI.transfer(0x00); 
  SPI.transfer(0x00);
  digitalWrite(pinCSLED, HIGH);
  
  //spegni tutti i led
  digitalWrite(pinCSLED, LOW);
  SPI.transfer(REGISTER_WRITE);
  SPI.transfer(GPIOA);
  SPI.transfer(0xFF); 
  SPI.transfer(0xFF);
  digitalWrite(pinCSLED, HIGH);
  
  // configuriamo in modo che l'interrupt venga cancellato nel momento in cui 
  // andiamo a leggere il registro INTCAP 
  // inoltre usiamo solo un interrupt per entrambe le porte
  digitalWrite(pinCSLED, HIGH);
  digitalWrite(pinCS, LOW);
  SPI.transfer(REGISTER_WRITE);
  SPI.transfer(IOCON);
  SPI.transfer(IOCON_INTCC | IOCON_MIRROR);
  digitalWrite(pinCS, HIGH);
    
  // Gli indirizzi dei registri delle porte A/B sono consecutivi
  for (byte i=0; i<2;i++) {
    //  tutti i pin come input
    digitalWrite(pinCS, LOW);
    SPI.transfer(REGISTER_WRITE);
    SPI.transfer(IODIRA + i);
    SPI.transfer(0xFF); 
    digitalWrite(pinCS, HIGH);
    
    //usiamo la logica inversa visto che si puo`
    digitalWrite(pinCS, LOW);
    SPI.transfer(REGISTER_WRITE);
    SPI.transfer(IPOLA + i);
    SPI.transfer(0xFF); 
    digitalWrite(pinCS, HIGH);
    
    //attiviamo le resistenze di pullup interne
    //pin flottanti sono HIGH!!
    digitalWrite(pinCS, LOW);
    SPI.transfer(REGISTER_WRITE);
    SPI.transfer(GPPUA + i);
    SPI.transfer(0xFF); 
    digitalWrite(pinCS, HIGH);
    
    //vogliamo che l'interrupt venga generato quando uno degli input divenda LOW
    //Quindi impostiamo i bit del registro DEFVAL tutti a 0
    digitalWrite(pinCS, LOW);
    SPI.transfer(REGISTER_WRITE);
    SPI.transfer(DEFVALA + i);
    SPI.transfer(0x00);
    digitalWrite(pinCS, HIGH);
    //impostiamo che l'interrupt venga generato se GPIO e` diverso da DEFVAL
    digitalWrite(pinCS, LOW);
    SPI.transfer(REGISTER_WRITE);
    SPI.transfer(INTCONA + i);
    SPI.transfer(0xFF); 
    digitalWrite(pinCS, HIGH);
    //attiviamo l'interrupt 
    digitalWrite(pinCS, LOW);
    SPI.transfer(REGISTER_WRITE);
    SPI.transfer(GPINTENA + i);
    SPI.transfer(0xFF);
    digitalWrite(pinCS, HIGH);
  
  }
  
  
  //inizializza midi e invia program change 0
  MIDI.begin();
  MIDI.sendProgramChange(0,midichannel);

  setLed(0);
  lcd.clear();
  lcd.print("Preset: ");
  lcd.print(program);
  
  // azzera e abilita INT0, NB: e` sul pin D2, non sul pin D0!!!!
  time = millis();
  EIFR = bit (INTF0);
  attachInterrupt(0, hoy, LOW); 
  
  
}


void loop(){
  
  // E` stato premuto un pulsante?
  if (interruptChange) {
    
    //quali pulsanti sono stati premuti?
    for (byte i=0;i<2;i++) { 
      digitalWrite(pinCS, LOW);
      SPI.transfer(REGISTER_READ);
      SPI.transfer(INTFA + i);
      ioexp[i]=SPI.transfer(0x00);
      digitalWrite(pinCS, HIGH);   
    }
    
    // devo leggere INTCAP per cancellare l'interrupt
    for (byte i=0;i<2;i++) { 
      digitalWrite(pinCS, LOW);
      SPI.transfer(REGISTER_READ);
      SPI.transfer(INTCAPA + i);
      SPI.transfer(0x00);
      digitalWrite(pinCS, HIGH);   
    }  
  
    //se e` stato premuto solo un pulsante ed e` passato tempo sufficiente
    //l'ultima condizione e` per scartare pulsanti non validi (per testing)
    if ( ( popCount(ioexp[0])+popCount(ioexp[1])==1 ) && ( millis()-time > wait_time ) && ( ioexp[1] < 9 ) ) {

      //stiamo cambiano decina?
      switch (ioexp[1]) {
        case B00000100:
          if (pgm_dec > 0) {
            pgm_dec--;
          }
          break;
        case B00001000:
          if (pgm_dec < 12) {
            pgm_dec++;
          }
          break;
        default:
          //se non stiamo cambiando decina allora...
          //architettura 8bit
          data = (word *)ioexp;
          pgm_uni = log2(*data);
          break;
      } 
      
      program = pgm_dec*10 + pgm_uni;
      //program change 0..127
      if (program > 127) {
        program = 127;
        pgm_uni = 7;
        pgm_dec = 12;
      }

      MIDI.sendProgramChange(program,midichannel);
      
      setLed(pgm_uni);
      lcd.clear();
      lcd.print("Preset: ");
      lcd.print(program);
      
      
      time = millis();
      
    }
    
    interruptChange = false;
    EIFR = bit (INTF0);   //AZZERA INT0 FLAG
    attachInterrupt(0, hoy, LOW); // interrupt elaborato, ora possiamo riabilitare.
  }
  
/*  set_sleep_mode (SLEEP_MODE_PWR_DOWN);   
  noInterrupts ();          // make sure we don't get interrupted before we sleep
  sleep_enable ();          // enables the sleep bit in the mcucr register
  attachInterrupt (0, hoy, LOW);  // wake up on low level
  interrupts ();           // interrupts allowed now, next instruction WILL be executed
  sleep_cpu ();            // here the device is put to sleep
 */
}



